/*
**  Document.m
**
**  Copyright (c) 2003
**
**  Author: Yen-Ju Chen <yjchenx@hotmail.com>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Document.h"
#include "Utilities.h"
#include "ToolbarView.h"
#include "FindManager.h"
#include "Inspector.h"
#include "MethodViewer.h"
#include "ScriptManager.h"
#include "DocumentController.h"
#include <CodeEditorView/CodeEditorView.h>
#include <CodeEditorView/CEView+Inspector.h>
#include <AppKit/AppKit.h>

static unsigned untitledCount = 0;

@interface Document (CodeEditorPrivate)
/* NSScrollView is autoreleased !! */
- (NSScrollView *) _scrollViewWithCodeEditorView: (NSRect) frame;
@end

@implementation Document (CodeEditorPrivate)
- (NSScrollView *) _scrollViewWithCodeEditorView: (NSRect) frame
{
  NSScrollView *scrollView;
  CodeEditorView *ceView;
  Class CEViewClass = CodeEditorViewClass();

  NSAssert(CEViewClass, @"No CodeEditorViewClass installed");

  scrollView = [[NSScrollView alloc] initWithFrame: frame];
  [scrollView setBorderType: NSBezelBorder];
  [scrollView setHasHorizontalScroller: YES];
  [scrollView setHasVerticalScroller: YES];
  [scrollView setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];
/*
  [[scrollView contentView] setAutoresizingMask: NSViewHeightSizable 
                            | NSViewWidthSizable];
*/
  [[scrollView contentView] setAutoresizesSubviews:YES];

  frame = [[scrollView contentView] bounds];
  ceView = [[CEViewClass alloc] initWithFrame: frame];
  [ceView setMinSize: NSMakeSize(0,0)];
  [ceView setMaxSize: NSMakeSize(1e7, 1e7)];
  [ceView setRichText: NO];
  [ceView setSelectable: YES];
  [ceView setVerticallyResizable: YES];
  [ceView setHorizontallyResizable: YES];
  [ceView setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];

  [scrollView setDocumentView: ceView];
  RELEASE(ceView);
  return AUTORELEASE(scrollView);
}
@end

@implementation Document

/* Document */

- (BOOL) close: (id) sender
{
  /* Go through all the CodeEditorView */
  while ([ceViews count] > 0)
    {
      if ([ceViews count] > 1)
        {
          [tabView selectFirstTabViewItem: self];
          ASSIGN(currentCEView, [(NSScrollView *)[[tabView selectedTabViewItem] view] documentView]);
        }
      if ([self closeTab: sender] == NO)
        {
          return NO;
        }
    }
  return YES;
}

- (BOOL) closeTab: (id) sender
{
  int result;

  if ([currentCEView isDocumentEdited])
    {
      result = NSRunAlertPanel(@"Unsaved document",
                               @"This document is edited and unsaved. "
                               @"Close it anyway ?",
                               @"No", @"Save and close", @"Yes", nil);
      switch(result)
        {
          case NSAlertDefaultReturn:
            return NO;
          case NSAlertAlternateReturn:
            [self saveDocument: self];
            break;
          case NSAlertOtherReturn:
            break;
        }
    }
  /* Handle tab close. 
   * Be careful about currentCEView, window title, and remove tab view is needed.
   * Remove files and ceViews.
   */
  if ([tabView superview]) /* attached */
    {
      int number = [tabView indexOfTabViewItem: [tabView selectedTabViewItem]];
      NSAssert([tabView numberOfTabViewItems] > 1, @"TabView shouldn't contain only one tag");

      if ([tabView numberOfTabViewItems] == 2) /* Put the rest tab into window */
        {
          id object;
          NSRect frame;
          /* Take of the one we don't want */
          [tabView removeTabViewItem: [tabView tabViewItemAtIndex: number]];
          /* Keep the one we want */
          object = [[tabView tabViewItemAtIndex: 0] view];
          RETAIN(object);
          /* Remove NSScrollView from tab view */
          [tabView removeTabViewItem: [tabView tabViewItemAtIndex: 0]];
          NSAssert(object, @"\nNSScrollView should exist");
          /* Set up right size */
          frame = [tabView frame];
          [object setFrame: frame];
          /* Take of tab view */
          RETAIN(tabView);
          [tabView removeFromSuperview];
          /* Put NSScrollView back to window */
          [[window contentView] addSubview: object];
          RELEASE(object);
          ASSIGN(currentCEView, [(NSScrollView *)object documentView]);
        }
      else /* Simply remove the tab */
        {
          int select = (number == 0) ? 0 : (number - 1);
          [tabView selectTabViewItemAtIndex: select];
          ASSIGN(currentCEView, [(NSScrollView *)[[tabView selectedTabViewItem] view] documentView]);
          [tabView removeTabViewItem: [tabView tabViewItemAtIndex: number]];
        }
      [files removeObjectAtIndex: number];
      [ceViews removeObjectAtIndex: number];
      NSAssert([files count] == [ceViews count], @"\nFiles should sync with cdViews");
      [self syncWindowTitle];   
    }
  else /* single CodeEditorView. Close the whole window */
    {
      [files removeAllObjects];
      [ceViews removeAllObjects];
      [self close];
    }
  return YES;
}

- (void) readFromFileInTab: (NSString *) filename ofType: (NSString *) type
{
  NSTabViewItem *tabItem;

  [self newDocumentInTab: self];

  tabItem = [tabView selectedTabViewItem];

  /* the NSTabViewItem should be the last one */
  NSAssert(([tabView numberOfTabViewItems]-1) == [tabView indexOfTabViewItem: tabItem], @"\nThe open file should go to the last tab view");
  NSAssert(currentCEView == [(NSScrollView *)[tabItem view] documentView], @"\ncurrentCEView should be the CodeEditorView in last tab");

  [files replaceObjectAtIndex: [tabView numberOfTabViewItems] -1
                   withObject: filename];
  [currentCEView loadFileNamed: filename];
  [currentCEView setLanguageName: type];
  [tabItem setLabel: [filename lastPathComponent]];

  /* Can use notification */
  [[Inspector sharedInspector] setCodeEditorView: currentCEView];
  [[FindManager sharedFindManager] setCodeEditorView: currentCEView];
  [[MethodViewer sharedMethodViewer] setCodeEditorView: currentCEView];
  [[ScriptManager sharedScriptManager] setCodeEditorView: currentCEView];

  [self syncWindowTitle];
  /* Not necessary */
  [self setFileType: type];
}

/* This method is also used by openDocumentInTab:,
 * which will open a new document in tab, then set the file
 */
- (void) newDocumentInTab: (id) sender
{
  id object;
  NSString *name;
  NSView *view;
  NSRect frame;

  if ([files count] == 1) // Put on tabView 
    {
      /* Toolbar may exist */
      NSEnumerator *e = [[[window contentView] subviews] objectEnumerator];
      while ((object = [e nextObject]))
        {
          if ([object isKindOfClass: [NSScrollView class]])
            {
              /* Move NSScrollView into NSTabView */
              name = [files objectAtIndex: 0];
              if ([name isEqualToString: @""])
                {  // No name
                  name = [NSString stringWithFormat: @"Untitled-%d", untitledCount++];
                }
              view = object;
              RETAIN(view); 
              [object removeFromSuperview];
              frame = [view frame];
              [tabView setFrame: frame];
              object = [[NSTabViewItem alloc] initWithIdentifier: name];
              [object setView: view];
              [object setLabel: [name lastPathComponent]];
              [tabView addTabViewItem: object];
              [[window contentView] addSubview: tabView];
              RELEASE(view);
              RELEASE(object);
              RELEASE(tabView);

              break;
            }
        }
    }

  /* Add one more tab */
  name = [NSString stringWithFormat: @"Untitled-%d", untitledCount++];

  frame = [[[tabView tabViewItemAtIndex: 0] view] frame];
  view = [self _scrollViewWithCodeEditorView: frame];

  object = [[NSTabViewItem alloc] initWithIdentifier: name];
  [object setView: view];
  [object setLabel: name];
  [tabView addTabViewItem: object];
  RELEASE(object);
  [tabView selectLastTabViewItem: self];
  [tabView setNeedsDisplay: YES];

  /* Trace each view */
  object = [(NSScrollView *)view documentView];
  NSAssert([object isKindOfClass: CodeEditorViewClass()], @"\nNSScrollView should contain CodeEditorView");
  [ceViews addObject: object];
  [files addObject: @""];
  ASSIGN(currentCEView, object);
  [window makeFirstResponder: currentCEView];

  NSAssert([files count] == [ceViews count], @"\nfiles should sync with ceViews");
}

- (void)saveDocument:(id)sender
{
  /* Save current CEView */
  NSString *filename = [currentCEView documentName];

  if ((filename == nil) || ([filename isEqualToString: @""]))
    {
      [self saveDocumentAs: sender];
      return;
    }

  [self writeWithBackupToFile: filename 
        ofType: [self fileType]
        saveOperation: NSSaveOperation];
}

- (void)saveDocumentAs:(id)sender
{
  [super saveDocumentAs: sender];
  /* Update tab label */
  if (tabView)
    {
      [[tabView selectedTabViewItem] setLabel: [[currentCEView documentName] lastPathComponent]];
      [tabView setNeedsDisplay: YES];
    }
}

- (void)saveDocumentTo:(id)sender
{
  [super saveDocumentTo: sender];
  if (tabView)
    {
      [[tabView selectedTabViewItem] setLabel: [[currentCEView documentName] lastPathComponent]];
      [tabView setNeedsDisplay: YES];
    }
}

- (void) revertDocumentToSaved: (id) sender
{
  if ([currentCEView documentName])
    {
      [currentCEView loadFileNamed: [currentCEView documentName]];

      /* Can use notification */
      [[Inspector sharedInspector] setCodeEditorView: currentCEView];
      [[FindManager sharedFindManager] setCodeEditorView: currentCEView];
      [[MethodViewer sharedMethodViewer] setCodeEditorView: currentCEView];
      [[ScriptManager sharedScriptManager] setCodeEditorView: currentCEView];
    }
}

- (BOOL)writeToFile:(NSString *)fileName ofType:(NSString *)type
{
  /* Don't care about the type, which might be incorrect */
  return [currentCEView saveFileAs: fileName];
}

- (BOOL)readFromFile:(NSString *)fileName ofType:(NSString *)type
{
  [files addObject: fileName]; // Track each file
  [self setFileType: type]; /* Don't need to care type actually */
  return YES;
}

- (void) setFileName: (NSString *) fileName
{
  /* Since Document is only a collection of files,
   * use random string as file name
   */
  [super setFileName: [[NSProcessInfo processInfo] globallyUniqueString]];
  [self syncWindowTitle];
}

- (CodeEditorView *) currentCodeEditorView
{
  return currentCEView;
}

/* TabView Delegate */
- (void) tabView: (NSTabView *) tabView 
         didSelectTabViewItem: (NSTabViewItem *) tabViewItem
{
  id object = [(NSScrollView *)[tabViewItem view] documentView];
  NSString *filename;

  NSAssert([object isKindOfClass: CodeEditorViewClass()], @"\ntabViewItem should contain CodeEditorView");

  ASSIGN(currentCEView, object);

  /* Can use notification */
  [[Inspector sharedInspector] setCodeEditorView: currentCEView];
  [[FindManager sharedFindManager] setCodeEditorView: currentCEView];
  [[MethodViewer sharedMethodViewer] setCodeEditorView: currentCEView];
  [[ScriptManager sharedScriptManager] setCodeEditorView: currentCEView];

  filename = [currentCEView documentName];
  if (filename)
    {
      [self setFileType: [documentController typeFromFileExtension: [filename pathExtension]]];
    }
  [self syncWindowTitle];
  [window makeFirstResponder: currentCEView];
}

/* Window Delegate */
- (BOOL) windowShouldClose: (NSNotification *) not
{
  return [self close: self];
}

/* Window Controller */
- (void) showFindPanel: (id) sender
{
  [[FindManager sharedFindManager] orderFrontFindPanel: sender];
}

- (void) showInspector: (id) sender
{
  [[Inspector sharedInspector] orderFrontInspectorPanel: sender];
}

- (void) showMethodViewer: (id) sender
{
  [[MethodViewer sharedMethodViewer] orderFrontMethodPanel: sender];
}

- (void) syncWindowTitle
{
  if ([currentCEView documentName] == nil)
    {
      if (tabView)
        [window setTitle: [[tabView selectedTabViewItem] label]];
      else
        [window setTitle: [NSString stringWithFormat: @"Untitled-%d", untitledCount++]];
    }
  else
    {
      NSString *filename = [[currentCEView documentName] lastPathComponent];
      NSString *path = [[currentCEView documentName] stringByDeletingLastPathComponent];
      if (path)
        [window setTitle: [NSString stringWithFormat: @"%@ -- %@", filename, path]];
      else
        [window setTitle: filename];
    }
}

- (void) refontify: (id) sender
{
  [currentCEView fontification];
}

- (void) markCurrentBlock: (id) sender
{
  [currentCEView markBlockAroundCursor];
}

- (void) search: (id) sender
{
  BOOL result = [currentCEView searchString: [searchField stringValue]
                                   backward: NO
                            caseInsensitive: NO];
  if (result == NO)
    {
      NSBeep();
      [searchField selectText: sender];
    }
}

- (void) goto: (id) sender
{
  int line = [sender intValue];
  if (line < 1)
    {
      NSBeep();
      NSRunAlertPanel(@"Tab spaces", @"Spaces for Tab key can't less than 0",
                      @"O.K.", nil, nil);
      [sender selectText: sender];
      return;
    }
  [sender setIntValue: line];

  [currentCEView gotoLineNum: line charInLine: 0];
  [window makeFirstResponder: currentCEView];
  [window makeKeyAndOrderFront: self];
}

- (void) windowDidBecomeKey: (NSNotification *) not
{
  /* Can use notification */
  [[Inspector sharedInspector] setCodeEditorView: currentCEView];
  [[FindManager sharedFindManager] setCodeEditorView: currentCEView];
  [[MethodViewer sharedMethodViewer] setCodeEditorView: currentCEView];
  [[ScriptManager sharedScriptManager] setCodeEditorView: currentCEView];
}

- (NSString *) windowNibName
{
  return @"Window.gorm";
}

- (void) windowControllerDidLoadNib: (NSWindowController *) controller
{
  NSScrollView *scrollView;
  CodeEditorView *ceView;
  NSRect frame;

  [super windowControllerDidLoadNib: controller];

  window = [controller window];
  /* CodeEditorView */

  frame = [[window contentView] bounds];
  scrollView = [self _scrollViewWithCodeEditorView: frame];
  [[window contentView] addSubview: scrollView];

  // Trace each view
  ceView = [scrollView documentView];
  [ceViews addObject: ceView];

  /* Files count should be 0 or 1 depending on 
   * whether newDocument: or openDocument:
   */
  if ([files count] == 1) // Load file for new window
    {
      [ceView loadFileNamed: [files objectAtIndex: 0]];
      [ceView setLanguageName: [documentController typeFromFileExtension: [[files objectAtIndex: 0] pathExtension]]];
    }
  else // New document
    {
      [files addObject: @""];
    }
  NSAssert([files count] == 1, @"\nOpen more than one file");
  NSAssert([files count] == [ceViews count], @"\nFiles should sync with ceViews");

  ASSIGN(currentCEView, ceView);

  [window makeFirstResponder: ceView];
  [self setFileName: [files objectAtIndex: 0]];
//  [window setDelegate: self];

  /* Toolbar */
  frame = [[window contentView] bounds];

  frame = NSMakeRect(0, frame.size.height-ToolbarHeight,
                     frame.size.width, ToolbarHeight);
  toolbar = [[ToolbarView alloc] initWithFrame: frame];
  [toolbar setAutoresizingMask: NSViewMinYMargin | NSViewWidthSizable];
  searchButton = [toolbar searchButton];
  searchField = [toolbar searchField];
  gotoField = [toolbar gotoField];

  [searchButton setTarget: self];
  [searchButton setAction: @selector(search:)];
  [searchField setTarget: self];
  [searchField setAction: @selector(search:)];
  [gotoField setTarget: self];
  [gotoField setAction: @selector(goto:)];

  /* TabView */
  frame = [[window contentView] bounds];
  tabView = [[NSTabView alloc] initWithFrame: frame];
  [tabView setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];
  [tabView setDelegate: self];
}

- (void) showSideView: (id) sender
{
  [currentCEView showSideTextView];
}

- (void) showToolbarView: (id) sender
{
  NSRect frame, bounds;
  NSView *view = nil;
  NSArray *array = [[window contentView] subviews];

  if ([array count] == 1) // Only NSScrollView or NSTabView
    { 
      view = [array objectAtIndex: 0];
      RETAIN(view);
      [view removeFromSuperview];

      bounds = [[window contentView] bounds];
      frame = NSMakeRect(0, bounds.size.height-ToolbarHeight,
                         bounds.size.width, ToolbarHeight);
      [toolbar setFrame: frame];
      [[window contentView] addSubview: toolbar];
      RELEASE(toolbar);

      frame = NSMakeRect(0, 0, bounds.size.width, 
                         bounds.size.height-ToolbarHeight);
      [view setFrame: frame];
      [[window contentView] addSubview: view];
      [window makeFirstResponder: searchField];
      RELEASE(view);
    }
  else if ([array count] == 2) // Toolbar and NSScrollView/NSTabView
    {
      NSEnumerator *e = [array objectEnumerator];
      id object;

      while ((object = [e nextObject]))
        {
          if (([object isKindOfClass: [NSScrollView class]]) ||
              ([object isKindOfClass: [NSTabView class]]))
            {
              view = object;
            }
          RETAIN(object);
          [object removeFromSuperview];
        }
      bounds = [[window contentView] bounds];
      [view setFrame: bounds];
      [[window contentView] addSubview: view];
      [window makeFirstResponder: currentCEView];
      RELEASE(view);
    }
}

- (id) init
{
  self = [super init];
  files = [NSMutableArray new];
  ceViews = [NSMutableArray new];
  documentController = [DocumentController sharedDocumentController];
  return self;
}

- (void) dealloc
{
  RELEASE(tabView);
  RELEASE(toolbar);
  RELEASE(currentCEView);
  RELEASE(ceViews);
  RELEASE(files);
  [super dealloc];
}

@end
