/*
**  FontAttributesManager.m
**
**  Copyright (c) 2003
**
**  Author: Yen-Ju Chen <yjchenx@hotmail.com>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "FontAttributesManager.h"
#include "Utilities.h"
#include <CodeEditorView/RulesetManager.h>
#include <AppKit/AppKit.h>

static FontAttributesManager *sharedManager;

@implementation FontAttributesManager

NSString *NSStringFromFont (NSFont *font)
{
  NSString *trait;
  NSFontTraitMask mask = [[NSFontManager sharedFontManager] traitsOfFont: font];
  NSNumber *size = [NSNumber numberWithFloat: [font pointSize]];

  if (mask & NSBoldFontMask)
    trait = @"Bold";
  else
    trait = @"Regular";

  if (mask & NSItalicFontMask)
    {
      if ([trait isEqualToString: @"Bold"])
        trait = [trait stringByAppendingString: @" Italic"];
      else
        trait = @"Italic";
    }

  return [NSString stringWithFormat: @"%@ %@ %d", 
                       [font displayName], trait, [size intValue]];
}

+ (FontAttributesManager *) sharedFontAttributesManager
{
  if (!sharedManager)
    {
      sharedManager = [[FontAttributesManager alloc] init];
    }
  return sharedManager;
}

- (NSPanel *) fontAttributesPanel;
{
  if (fontAttributesPanel == nil)
    {
      if(NO == [NSBundle loadNibNamed: @"FontAttributesPanel.gorm" owner: self])
        {
           NSBeep();
           return nil;
        }
    }
  return fontAttributesPanel;
}

/* Font Attributes Panel */
- (void) orderFrontFontAttributesPanel: (id) sender
{
  if ([self fontAttributesPanel] == nil)
    {
       NSBeep();
       return;
    }

  /* Set up attributes */
  rulesetManager = [RulesetManagerClass() sharedRulesetManager];

  /* Cache color */
  [colorCache setObject: [rulesetManager colorForType: @"Normal"]
                 forKey: @"Normal"];
  [colorCache setObject: [rulesetManager colorForType: @"Comments"]  
                 forKey: @"Comments"]; 
  [colorCache setObject: [rulesetManager colorForType: @"Strings"]  
                 forKey: @"Strings"]; 
  [colorCache setObject: [rulesetManager colorForType: @"Preprocessor"]  
                 forKey: @"Preprocessor"]; 
  [colorCache setObject: [rulesetManager colorForType: @"Keywords"]  
                 forKey: @"Keywords"]; 
  [colorCache setObject: [rulesetManager colorForType: @"KnownTypes"]  
                 forKey: @"KnownTypes"]; 

  /* Cache Font */
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"Normal"]
                forKey: @"Normal"];
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"Comments"]
                forKey: @"Comments"];
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"Strings"]
                forKey: @"Strings"];
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"Preprocessor"]
                forKey: @"Preprocessor"];
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"Keywords"]
                forKey: @"Keywords"];
  [fontCache setObject: [rulesetManager convertFont: nil toType: @"KnownTypes"]
                forKey: @"KnownTypes"];

  /* Get everything from defaults */
  [self revertToSavedFontAttributes: self];


  [fontAttributesPanel makeKeyAndOrderFront: self];
}

- (void) colorAction: (id) sender
{
  if (sender == normalColor)
    {
      [rulesetManager setColor: [sender color] forType: @"Normal"];
      return;
    }
  if (sender == commentColor)
    {
      [rulesetManager setColor: [sender color] forType: @"Comments"];
      return;
    }
  if (sender == stringColor)
    {
      [rulesetManager setColor: [sender color] forType: @"Strings"];
      return;
    }
  if (sender == preprocessorColor)
    {
      [rulesetManager setColor: [sender color] forType: @"Preprocessor"];
      return;
    }
  if (sender == keywordColor)
    {
      [rulesetManager setColor: [sender color] forType: @"Keywords"];
      [rulesetManager setColor: [sender color] forType: @"ExtraKeywords"];
      return;
    }
  if (sender == typeColor)
    {
      [rulesetManager setColor: [sender color] forType: @"KnownTypes"];
      [rulesetManager setColor: [sender color] forType: @"ExtraKnownTypes"];
      return;
    }
}

- (void) changeFont: (id) sender
{
  NSFont *font = [fontManager convertFont: [fontCache objectForKey: selectedFont]];

  [rulesetManager setFont: font forType: selectedFont];

  if ([selectedFont isEqualToString: @"Normal"])
    {
      [normalFont setTitle: NSStringFromFont(font)];
      return;
    }
  if ([selectedFont isEqualToString: @"Comments"])
    {
      [commentFont setTitle: NSStringFromFont(font)];
      return;
    }
  if ([selectedFont isEqualToString: @"Strings"])
    {
      [stringFont setTitle: NSStringFromFont(font)];
      return;
    }
  if ([selectedFont isEqualToString: @"Preprocessor"])
    {
      [preprocessorFont setTitle: NSStringFromFont(font)];
      return;
    }
  if ([selectedFont isEqualToString: @"Keywords"])
    {
      [keywordFont setTitle: NSStringFromFont(font)];
      return;
    }
  if ([selectedFont isEqualToString: @"KnownTypes"])
    {
      [typeFont setTitle: NSStringFromFont(font)];
      return;
    }
}

- (void) fontAction: (id) sender
{
  [fontManager orderFrontFontPanel: self];
  [fontAttributesPanel makeKeyAndOrderFront: self];

  NSLog(@"fontAction %@", sender);

  if (sender == normalFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"Normal"] 
                        isMultiple: NO];
      selectedFont = @"Normal";
      return;
    }
  if (sender == commentFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"Comments"]
                        isMultiple: NO];
      selectedFont = @"Comments";
      return;
    }
  if (sender == stringFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"Strings"]
                        isMultiple: NO];
      selectedFont = @"Strings";
      return;
    }
  if (sender == preprocessorFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"Preprocessor"]
                        isMultiple: NO];
      selectedFont = @"Preprocessor";
      return;
    }
  if (sender == keywordFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"Keywords"]
                        isMultiple: NO];
      selectedFont = @"Keywords";
      return;
    }
  if (sender == typeFont)
    {
      [fontManager setSelectedFont: [fontCache objectForKey: @"KnownTypes"]
                        isMultiple: NO];
      selectedFont = @"KnownTypes";
      return;
   }
}

- (void) revertToSavedFontAttributes: (id) sender
{
  [normalColor setColor: [colorCache objectForKey: @"Normal"]];
  [commentColor setColor: [colorCache objectForKey: @"Comments"]];
  [stringColor setColor: [colorCache objectForKey: @"Strings"]];
  [preprocessorColor setColor: [colorCache objectForKey: @"Preprocessor"]];
  [keywordColor setColor: [colorCache objectForKey: @"Keywords"]];
  [typeColor setColor: [colorCache objectForKey: @"KnownTypes"]];
  [normalFont setTitle: NSStringFromFont([fontCache objectForKey: @"Normal"])];
  [commentFont setTitle: NSStringFromFont([fontCache objectForKey: @"Comments"])];
  [stringFont setTitle: NSStringFromFont([fontCache objectForKey: @"Strings"])];
  [preprocessorFont setTitle: NSStringFromFont([fontCache objectForKey: @"Preprocessor"])];
  [keywordFont setTitle: NSStringFromFont([fontCache objectForKey: @"Keywords"])];
  [typeFont setTitle: NSStringFromFont([fontCache objectForKey: @"KnownTypes"])];
}

- (void) windowWillClose: (NSNotification *) not
{
  if ([not object] == fontAttributesPanel)
    {
      [rulesetManager writeFontAttributesToUserDefaults];
      [[NSColorPanel sharedColorPanel] performClose: self];
      [[fontManager fontPanel: NO] performClose: self];
    }
}

- (id) init
{
  self = [super init];

  colorCache = [NSMutableDictionary new];
  fontCache = [NSMutableDictionary new];
  fontManager = [NSFontManager sharedFontManager];

  return self;
}

- (void) dealloc
{
  RELEASE(colorCache);
  RELEASE(fontCache);
  [super dealloc];
}

@end
