/*
**  CodeEditorView.m
**
**  Copyright (c) 2002, 2003
**
**  Author: Yen-Ju Chen <yjchenx@hotmail.com>
**          Bjoern Giesler <bjoern@giesler.de>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CodeEditorView.h"
#include "RulesetManager.h"
#include "CodeEditorViewPreference.h"
#include "CEView+BlockMarking.h"
#include "CEView+Indentation.h"
#include "CEView+TextUtils.h"
#include "CEView+Fontification.h"
#include "CEView+Inspector.h"
#include "NSAttributedString+FindAndReplace.h"
#include "BundleLoader.h"
#include "FontificationHandler.h"
#include <AppKit/AppKit.h>

@implementation CodeEditorView

- (void) showSideTextView
{
  if (hasSideTextView)
    {
      [self setTextContainerInset: NSMakeSize(0,0)];
      hasSideTextView = NO;
    }
  else
    {
      [self setTextContainerInset: NSMakeSize(40, 0)];
      hasSideTextView = YES;
    }
}

- (BOOL) isDocumentEdited
{
  return documentEdited;
}

- (void) setDocumentEdited: (BOOL) bool
{
  documentEdited = bool;
  if (bool == YES)
    [[self window] setDocumentEdited: YES];
}

- (void) fontification
{
  if ([self supportLanguage: languageName])
    {
       [self unmarkPreviouslyMarkedBlock];

       fontifying = YES;
       [self fontifyAccordingToLanguage: languageName];
       fontifying = NO; 

       [self markBlockAroundCursor];
    }
  needsFontification = NO;

}

- (id) initWithFrame: (NSRect) frame
{
  Class bundleClass;
  NSString *bundleName;
  int i, count;

  [super initWithFrame: frame];
  
  markedStartRange.location = NSNotFound;
  markedEndRange.location = NSNotFound;

  documentName = nil;
  languageName = nil;
  needsFontification = NO;
  hasBlockMarked = NO;

  [self setString: @""];
  [self setSelectedRange: NSMakeRange(0, 0)];
  [self setRichText: NO];
  [self setImportsGraphics: NO];
  [self setUsesFontPanel: NO];
  [self setUsesRuler: NO];
  
  [self performSelector: @selector(periodicTask:)
	withObject: self
	afterDelay: .2];

  [[self textStorage] setDelegate: self];

  preference = [CodeEditorViewPreference sharedCodeEditorViewPreference];
  
  hasSideTextView = NO;

  if ([preference displaySideView])
    {
      [self showSideTextView];
    }

  [[self textContainer] setContainerSize: NSMakeSize(1e7, 1e7)];

  /* Load language bundle */
  bundleLanguages = [NSMutableArray new];
  bundleNames = [[BundleLoader defaultLoader] allNamesOfBundles];
  RETAIN(bundleNames);

  count = [bundleNames count];
 
  for (i = 0; i < count; i++)
    {
      bundleName = [bundleNames objectAtIndex: i];
      bundleClass = [[BundleLoader defaultLoader] classNameOf: bundleName];
      [bundleLanguages addObject: [bundleClass language]];
    }

  return self;
}

- (BOOL)loadFileNamed: (NSString *)filename;
{
  NSString *str;
  NSFont *font;

  str = [NSString stringWithContentsOfFile: filename];
  if(!str)
    {
      NSLog(@"Couldn't read file.");
      return NO;
    }

  fontifying = YES;
  font = [[RulesetManager sharedRulesetManager] convertFont: nil 
                                                     toType: @"Normal"];
  ASSIGN(normalFont, [font copy]);

  [self setString: str];

  [[self textStorage] setAttributes: [NSDictionary dictionaryWithObject: font
                                                   forKey: NSFontAttributeName]
                              range: NSMakeRange(0, [str length])];

  [self setSelectedRange: NSMakeRange(0,0)];

  ASSIGN(documentName, filename);

  fontifying = NO;

  return YES;
}

- (BOOL)saveFileAs: (NSString *)filename
{
  BOOL saveOK = NO;

  if(![[filename pathExtension] compare: @"rtf"])
    saveOK = [[[self textStorage]
		RTFFromRange: NSMakeRange(0, [[self string] length])
		documentAttributes: nil]
	       writeToFile: filename atomically: NO];
  else
    saveOK = [[self string] writeToFile: filename atomically: NO];

  if(!saveOK)
    return NO;

  ASSIGN(documentName, filename);
  
  [self setDocumentEdited: NO];

  [[self window] setTitleWithRepresentedFilename: filename];
  [[self window] setDocumentEdited: NO];
  
  return YES;
}

- (BOOL)saveFile
{
  id panel, filename;
  int result;
  
  if(!documentName)
    {
      panel = [NSSavePanel savePanel];
      result = [panel runModal];
      if(result != NSOKButton)
	return NO;
  
      filename = [panel filename];
    }
  else
    filename = documentName;

  return [self saveFileAs: filename];
}

- (NSString *)documentName
{
  return documentName;
}

- (NSString *)languageName
{
  if(!languageName)
    return @"Text";
  return languageName;
}

- (void)setLanguageName: (NSString *) name
{
  NSString *bundleName;
  unsigned int index;
  ASSIGN(languageName, name);
  needsFontification = YES;

  /* Get block handler */
  index = [bundleLanguages indexOfObject: name];
  if (index != NSNotFound)
    {
      bundleName = [[[BundleLoader defaultLoader] allNamesOfBundles] objectAtIndex: index];
      fontificationHandlerClass = [[BundleLoader defaultLoader] classNameOf: bundleName];
      blockHandlerClass = [fontificationHandlerClass blockHandlerClass];
      indentationHandlerClass = [fontificationHandlerClass indentationHandlerClass];
    }
}

- (void)drawRect: (NSRect) rect
{
  [super drawRect: rect];

  if (hasSideTextView)
    {
      /* draw side text view */
      unsigned int i, len, totalLines = 0;
      NSString *text;
      NSRect glyphRect, visibleRect;
      NSString *s;
      NSDictionary *attributes;

      text = [[self textStorage] string];
      len = [text length];
      layoutManager = [self layoutManager];

      visibleRect = [[self superview] visibleRect];
      [[NSColor lightGrayColor] set];
      [NSBezierPath fillRect: NSMakeRect(visibleRect.origin.x, visibleRect.origin.y,
                                         40, visibleRect.size.height)];

      attributes = [NSDictionary dictionaryWithObjectsAndKeys:
                             [NSColor blackColor], NSForegroundColorAttributeName,
                             [NSFont boldSystemFontOfSize: 12], NSFontAttributeName,
                              nil];
      /* Draw marked area first */
      if (hasSideMarked)
        {
          [[NSColor yellowColor] set];
          [NSBezierPath fillRect: sideMarkedRect];
        }

      for (i = 0; i < len; i++)
        {
           if ([text characterAtIndex: i] == '\n')
             {
               totalLines++;
               glyphRect = [layoutManager lineFragmentRectForGlyphAtIndex: i
                                                           effectiveRange: NULL];
               // glyphRect.origin.y ++;
               visibleRect.origin.y -= 5;
               visibleRect.size.height += 10;
               if (NSPointInRect(glyphRect.origin, visibleRect))
                 {
                   s = [[NSString alloc] initWithFormat: @"%d", totalLines];
                   [s drawAtPoint: glyphRect.origin withAttributes: attributes];
                   RELEASE(s);
                 }
             }
        }
    }
}

- (void) finishedTypingWord: (NSString *) word
                      range: (NSRange) range
{
  if ([preference autoFontificationEachLine])
    {
      NSRange lineRange;
      NSFont *font;
      NSColor *color;

      lineRange = [self rangeOfLineAtCursor];

      /* Delete word */
      if (NSMaxRange(range) > [[self textStorage] length])
        range.length = [[self textStorage] length] - range.location;

      if ([self supportLanguage: languageName])
        {
          [self unmarkPreviouslyMarkedBlock];

          fontifying = YES;

          color = [[RulesetManager sharedRulesetManager] colorForType: @"Normal"];
          font = [[RulesetManager sharedRulesetManager] convertFont: nil
                                                        toType: @"Normal"];

          [[self textStorage] addAttribute: NSFontAttributeName
                                     value: font
                                     range: lineRange];
          [[self textStorage] addAttribute: NSForegroundColorAttributeName
                                     value: color
                                     range: lineRange];

          [self fontifyAccordingToLanguage: languageName inRange: lineRange];
          fontifying = NO; 
                    
          [self markBlockAroundCursor];
        }
      needsFontification = NO;
    }
}

- (void) keyDown: (NSEvent *) theEvent
{
  unichar fnkey;
  BOOL cursorMotion;
  NSRange wordAtBeginRange;
  NSString *wordAtBegin, *wordAtEnd;
  NSString *tabString;
  unsigned point;
  
  userPressingKey = YES;

  if (hasBlockMarked)
    {
      [self unmarkPreviouslyMarkedBlock];
      hasBlockMarked = NO;
    }

  wordAtBegin = [self wordAtCursor];
  wordAtBeginRange = [self rangeOfWordAtCursor];
  
  if([[theEvent characters] length])
    fnkey = [[theEvent characters] characterAtIndex: 0];
  else
    fnkey = (unichar)0;

  if((fnkey == NSLeftArrowFunctionKey) ||
     (fnkey == NSRightArrowFunctionKey) ||
     (fnkey == NSUpArrowFunctionKey) ||
     (fnkey == NSDownArrowFunctionKey))
    {
      cursorMotion = YES;

      [super keyDown: theEvent];
    }
  else
    {
      cursorMotion = NO;
      [self setDocumentEdited: YES];
    
      if([[theEvent characters] isEqualToString: @"\t"])
        {
          if ((tabString = [preference tabString]))
            [self insertText: tabString];
          else /* Auto Indentation */
            {
	      point = [self selectedRange].location;
	      point += [self syntacticallyIndentLineAtIndex: point];
	      [self setSelectedRange: NSMakeRange(point, 0)];
            }
        }
      else if([[theEvent characters] isEqualToString: @"\r"])
        {
          // Return
          [super keyDown: theEvent];
          if ([preference autoIndentation])
            {
              point = [self selectedRange].location;
              point += [self syntacticallyIndentLineAtIndex: point];
              [self setSelectedRange: NSMakeRange(point, 0)];
            }
        }
      else
        [super keyDown: theEvent];

      needsFontification = YES;
    }

    /* Mark delimiter */
    wordAtEnd = [self wordAtCursor];

    if ([preference autoMarkBlock])
      {
        if ([blockHandlerClass isDelimiter: wordAtEnd])
          [self markBlockAroundCursor];
      }
 
  if (cursorMotion == NO)
    {
      if (wordAtBegin) 
        {
          if([wordAtEnd rangeOfString: wordAtBegin].location == NSNotFound)
            [self finishedTypingWord: wordAtBegin
                               range: wordAtBeginRange];
        }
    }
  userPressingKey = NO;
}

- (void) markBlockAroundCursor
{
  [self findAndMarkBlockAtCursor];
  hasBlockMarked = YES;
}

- (void)periodicTask: (id) sender
{
  static float sinceFontify;
  static float sincePeriodicTask;
  static int lastLines;

  sinceFontify += .2;
  sincePeriodicTask += .2;

  if(userPressingKey)
    {
      /* Skip this one */
      sinceFontify = 0;
      sincePeriodicTask = 0;
      [self performSelector: @selector(periodicTask:)
	    withObject: self
	    afterDelay: .2];
      
      return;
    }
      
  if(sinceFontify > [preference fontificationInterval])
    {
      if([preference autoFontification] && needsFontification)
        {
          /* Not over threshold */
          if ( !([preference autoFontificationThreshold] &&
                 ([preference fontificationThreshold] < lastLines)) )
            {
              if ((documentEdited && [preference autoFontification]) ||
                  needsFontification)
                {
                  [self fontification];  
                 }
	      needsFontification = NO;
            }
        }
      sinceFontify = 0;
    }

  /* Check lines every 10 seconds in order to get accurate lines
   * so that it can apply to threshold */
  if (sincePeriodicTask > 10)
    {
      [self getNumberOfGlyphs: NULL
                        lines: &lastLines];
    }

  [self performSelector: @selector(periodicTask:)
	withObject: self
	afterDelay: .2];
}

- (BOOL) searchString: (NSString *) string
             backward: (BOOL) backward
      caseInsensitive: (BOOL) caseInsensitive
{
  NSRange cursor, range;
  cursor = [self selectedRange];

  if (cursor.location == NSNotFound)
    cursor = NSMakeRange(0, [[self textStorage] length]);

  if (backward == NO)
    {
      range = [[self textStorage] rangeOfString: string
                         inRange: NSMakeRange(cursor.location + cursor.length,
                                              [[self textStorage] length] - cursor.location - cursor.length)
                        backward: backward
                 caseInsensitive: caseInsensitive];
    }
  else
    {
      range = [[self textStorage] rangeOfString: string
                         inRange: NSMakeRange(0, cursor.location)
                        backward: backward
                 caseInsensitive: caseInsensitive]; 
    }

  if (range.location == NSNotFound)
    return NO;  

  [self setSelectedRange: range];
  [self scrollRangeToVisible: range];
  return YES;
}

- (void)textStorageDidProcessEditing: (NSNotification *) not
{
  if(!fontifying)
    {
      /* [not object] is the whole attributed string */
      [self setDocumentEdited: YES];
    }
}

- (NSArray *) supportedLanguages
{
  return bundleLanguages;
} 

- (void) dealloc
{
  RELEASE(languageName);
  RELEASE(documentName);
  RELEASE(normalFont);
  RELEASE(bundleLanguages);
  RELEASE(bundleNames);
  [super dealloc];
}

@end
